/***************************************************************************

  output.c

  The object file creator

  (c) 2000-2001 Benot Minisini <gambas@freesurf.fr>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#define _OUTPUT_C

#include "common.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>

#include "error.h"
#include "memory.h"
#include "str.h"
#include "file.h"

#include "table.h"
#include "compile.h"
#include "magic.h"

#include "output.h"


/*#define DEBUG*/


PRIVATE FILE *File;

PRIVATE CLASS *Class;

PRIVATE long PosCheckSum;
PRIVATE long StringAddr;
PRIVATE TABLE *StringTable;


#define START_SECTION()  long pos = write_long(0);
#define END_SECTION()    write_pad(); write_long_at(pos, get_pos() - pos - sizeof(long));


PRIVATE void output_init(void)
{
  TABLE_create(&StringTable, sizeof(OUTPUT_SYMBOL), TF_NORMAL);
  StringAddr = 0;
}

PRIVATE void output_exit(void)
{
  TABLE_delete(&StringTable);
}

PRIVATE long get_string(char *string, int len)
{
  OUTPUT_SYMBOL *sym;

  if (len < 0)
    len = strlen(string);

  if (!TABLE_add_symbol(StringTable, string, len, (SYMBOL **)&sym, NULL))
  {
    sym->pos = StringAddr;
    StringAddr += len + 1;
  }

  #ifdef DEBUG
  printf("'%.*s' -> %ld\n", len, string, sym->pos);
  #endif

  return sym->pos;
}


PRIVATE long get_pos(void)
{
  return ftell(File);
}

PRIVATE void write_byte(unsigned char val)
{
  #ifdef DEBUG
  printf("%ld : b %u 0x%X\n", get_pos(), val, val);
  #endif

  /* MSB & LSB */
  if (fwrite(&val, sizeof(char), 1, File) != 1)
    THROW("Write error");
}

PRIVATE void write_short(ushort val)
{
  #ifdef DEBUG
  printf("%ld : i %u 0x%X\n", get_pos(), val, val);
  #endif

  /* MSB & LSB */
  if (fwrite(&val, sizeof(val), 1, File) != 1)
    THROW("Write error");
}

PRIVATE long write_long(ulong val)
{
  long pos = get_pos();

  #ifdef DEBUG
  printf("%ld : l %lu 0x%lX\n", get_pos(), val, val);
  #endif

  /* MSB & LSB */
  if (fwrite(&val, sizeof(val), 1, File) != 1)
    THROW("Write error");
  return pos;
}

PRIVATE void write_long_at(long pos, ulong val)
{
  long prev = get_pos();
  
  fseek(File, pos, SEEK_SET);

  #ifdef DEBUG
  printf("@ ");
  #endif

  write_long(val);

  fseek(File, prev, SEEK_SET);
}


PRIVATE void write_string(char *str, int len)
{
  #ifdef DEBUG
  printf("%ld : s \"%.*s\"\n", get_pos(), len, str);
  #endif

  if (fwrite(str, sizeof(char), len, File) != len)
    THROW("Write error");
  write_byte(0);
}


PRIVATE void write_buffer(void *str, long len)
{
  #ifdef DEBUG
  printf("%ld : buffer %ld octets\n", get_pos(), len);
  #endif

  if (len == 0)
    return;

  if (fwrite(str, sizeof(char), len, File) != len)
    THROW("Write error");
}


PRIVATE void write_pad(void)
{
  while (get_pos() & 0x1)
    write_byte(0);
}


PRIVATE void output_header(void)
{
  #ifdef DEBUG
  printf("Section #1 : Header\n");
  #endif

  /* magic */
  write_long(OUTPUT_MAGIC);
  /* version */
  write_long(OUTPUT_VERSION);
  /* checksum */
  PosCheckSum = write_long(0);
  /* flag */
  if (JOB->debug)
    write_long(1);
  else
    write_long(0);
}


PRIVATE void output_class(void)
{
  START_SECTION()

  #ifdef DEBUG
  printf("Section #2 : Classe\n");
  #endif

  /* classe parente */
  write_short(Class->parent);
  /* size_static */
  write_long(Class->size_stat);
  /* size_dynamic */
  write_long(Class->size_dyn);

#if 0
  /* desc_static */
  write_long(ARRAY_count(Class->stat));
  /* desc_dynamic */
  write_long(ARRAY_count(Class->dyn));
  /* desc_class */
  write_long(ARRAY_count(Class->class));
  /* desc_constant */
  write_long(ARRAY_count(Class->constant));
  /* desc_method */
  write_long(ARRAY_count(Class->function));
  /* desc_symbol */
  write_long(ARRAY_count(Class->table));
#endif

  END_SECTION()
}


PRIVATE void output_desc(void)
{
  START_SECTION()

  int i, n, nn = 0;
  CLASS_SYMBOL *csym;
  TYPE type;
  short out_type;

  n = TABLE_count(Class->table);

  #ifdef DEBUG
  printf("Section #3 : Description\n");
  #endif

  for (i = 0; i < n; i++)
  {
    csym = (CLASS_SYMBOL *)TABLE_get_symbol(Class->table, i);
    csym = (CLASS_SYMBOL *)TABLE_get_symbol(Class->table, csym->symbol.sort);

    type = csym->global.type;

    if (TYPE_is_public(type))
    {
      nn++;
      /* name */
      write_long(get_string(csym->symbol.name, csym->symbol.len));

      switch (TYPE_get_kind(type))
      {
        case TK_VARIABLE:

          /* type de donnes */
          write_long(TYPE_long(csym->global.type));
          /* param */
          write_long(csym->global.value);
          /* read */
          write_long(0);
          /* write */
          write_long(0);

          if (TYPE_is_static(type))
            out_type = CD_STATIC_PROPERTY_ID;
          else
            out_type = CD_PROPERTY_ID;

          break;

        case TK_CONST:

          /* type de donnes */
          write_long(TYPE_long(csym->global.type));
          /* param */
          write_long(csym->global.value);
          /* read */
          write_long(0);
          /* write */
          write_long(0);

          out_type = CD_CONSTANT_ID;

          break;

        case TK_FUNCTION:

          /* type de la valeur de retour */
          write_long(TYPE_long(csym->global.type));
          /* exec */
          write_long(csym->global.value);
          /* signature */
          write_long(0);
          /* nparam */
          write_long(0);

          if (TYPE_is_static(type))
            out_type = CD_STATIC_METHOD_ID;
          else
            out_type = CD_METHOD_ID;

          break;

        case TK_EVENT:

          /* type de la valeur de retour */
          write_long(TYPE_long(csym->global.type));
          /* exec */
          write_long(csym->global.value);
          /* signature */
          write_long(0);
          /* nparam */
          write_long(0);

          out_type = CD_EVENT_ID;

          break;

        default:
          ERROR_panic("output_desc: unknown symbol type");
          goto SUIVANT;
       }

      /* type de symbole */
      write_long(out_type);
    }

SUIVANT:
  }

  #ifdef DEBUG
  printf("Section #3 : Description (%d)\n", nn);
  #endif

  END_SECTION()
}



PRIVATE void output_constant()
{
  START_SECTION()

  int i, n;
  CONSTANT *constant;
  SYMBOL *sym;

  n = ARRAY_count(Class->constant);

  #ifdef DEBUG
  printf("Section #4 : Constantes (%d)\n", n);
  #endif

  for (i = 0; i < n; i++)
  {
    constant = &Class->constant[i];

    /* type */
    write_long(TYPE_long(constant->type));
    /* value */
    if (TYPE_can_be_long(constant->type))
    {
      write_long(constant->value);
      write_long(0);
    }
    else
    {
      sym = TABLE_get_symbol(TYPE_get_id(constant->type) == T_STRING ? Class->string : Class->table, constant->value);
      write_long(get_string(sym->name, sym->len));
      write_long(sym->len);
    }
  }

  END_SECTION()
}


PRIVATE void output_class_ref(void)
{
  START_SECTION()

  int i, n;
  SYMBOL *sym;

  n = ARRAY_count(Class->class);

  #ifdef DEBUG
  printf("Section #5 : Rfrences aux classes externes (%d)\n", n);
  #endif

  for (i = 0; i < n; i++)
  {
    sym = TABLE_get_symbol(Class->table, Class->class[i]);
    write_long(get_string(sym->name, sym->len));
  }

  END_SECTION()
}


PRIVATE void output_unknown_ref(void)
{
  START_SECTION()

  int i, n;
  SYMBOL *sym;

  n = ARRAY_count(Class->unknown);

  #ifdef DEBUG
  printf("Section #6 : Rfrences aux symboles externes (%d)\n", n);
  #endif

  for (i = 0; i < n; i++)
  {
    sym = TABLE_get_symbol(Class->table, Class->unknown[i]);
    write_long(get_string(sym->name, sym->len));
  }

  END_SECTION()
}


PRIVATE void output_static(void)
{
  START_SECTION()

  int i, n;
  VARIABLE *var;
  
  n = ARRAY_count(Class->stat);
  
  #ifdef DEBUG
  printf("Section #7 : variables statiques (%d)\n", n);
  #endif
  
  for (i = 0; i < n; i++)
  {
    var = &Class->stat[i];

    /* type */
    write_long(TYPE_long(var->type));
    /* addr */
    write_long(var->pos);
  }
  
  END_SECTION()
}


PRIVATE void output_dynamic(void)
{
  START_SECTION()
  
  int i, n;
  VARIABLE *var;
  
  n = ARRAY_count(Class->dyn);

  #ifdef DEBUG
  printf("Section #8 : variables dynamiques (%d)\n", n);
  #endif
  
  for (i = 0; i < n; i++)
  {
    var = &Class->dyn[i];

    /* type */
    write_long(TYPE_long(var->type));
    /* addr */
    write_long(var->pos);
  }

  END_SECTION()
}


PRIVATE void output_event(void)
{
  START_SECTION()

  int i, n;
  EVENT *event;
  SYMBOL *sym;

  n = ARRAY_count(Class->event);

  #ifdef DEBUG
  printf("Section #9 : Evnements (%d)\n", n);
  #endif

  for (i = 0; i < n; i++)
  {
    event = &Class->event[i];

    /* type */
    write_long(TYPE_long(event->type));
    /* n_param */
    write_short(event->nparam);
    /* desc_param */
    write_long(0);
    /* name */
    sym = TABLE_get_symbol(Class->table, event->name);
    write_long(get_string(sym->name, sym->len));
  }

  END_SECTION()
}


PRIVATE void output_method(void)
{
  START_SECTION()

  int i, n;
  FUNCTION *func;
  SYMBOL *sym;

  n = ARRAY_count(Class->function);

  #ifdef DEBUG
  printf("Section #10 : Mthodes (%d)\n", n);
  #endif

  for (i = 0; i < n; i++)
  {
    func = &Class->function[i];

    /* type */
    write_long(TYPE_long(func->type));
    /* n_param */
    write_short(func->nparam);
    /* n_param_min */
    write_short(func->npmin);
    /* n_local */
    write_short(func->nlocal);
    /* n_local */
    write_short(func->nctrl);
    /* stack_usage */
    write_short(func->stack);
    /* gestion d'erreur */
    if (func->catch && func->finally)
      write_short(Min(func->finally, func->catch));
    else if (func->catch)
      write_short(func->catch);
    else
      write_short(func->finally);
    /* addr_code */
    write_long(ARRAY_count(func->code));
    /* desc_param */
    write_long(0);
    /* desc_local */
    write_long(0);

    if (JOB->debug && func->pos_line != NULL)
    {
      /* line */
      write_short(func->line);
      write_short(ARRAY_count(func->pos_line));
      /* pos_line */
      write_long(0);
      /* nom */
      sym = TABLE_get_symbol(Class->table, func->name);
      write_long(get_string(sym->name, sym->len));
    }
    else
    {
      write_short(0);
      write_short(0);
      write_long(0);
      write_long(0);
    }
  }

  END_SECTION()
}


PRIVATE void output_param_local(void)
{
  START_SECTION()
  
  int i, j;
  FUNCTION *func;
  EVENT *event;
  PARAM *param;

  #ifdef DEBUG
  printf("Section #11 : Paramtres\n");
  #endif

  for (i = 0; i < ARRAY_count(Class->function); i++)
  {
    func = &Class->function[i];

    if (func->name != NO_SYMBOL)
    {
      /* Les paramtres sont remis dans les variables locales !
      for (j = 0; j < func->nparam; j++)
      {
        param = &func->param[j];
        write_long(param->type);
      }
      */

      for (j = 0; j < func->nlocal + func->nparam; j++)
      {
        param = &func->local[j];

        /* type */
        write_long(TYPE_long(param->type));
      }
    }
  }

  for (i = 0; i < ARRAY_count(Class->event); i++)
  {
    event = &Class->event[i];

    for (j = 0; j < event->nparam; j++)
    {
      param = &event->param[j];

      /* type */
      write_long(TYPE_long(param->type));
    }
  }

  END_SECTION()
}


PRIVATE void output_array(void)
{
  START_SECTION()

  long i, j, p;
  CLASS_ARRAY *array;

  #ifdef DEBUG
  printf("Section #13 : Chanes\n");
  #endif

  p = ARRAY_count(Class->array) * sizeof(long);

  for (i = 0; i < ARRAY_count(Class->array); i++)
  {
    array = &Class->array[i];
    write_long(p);
    p += sizeof(long) + array->ndim * sizeof(long);
  }

  for (i = 0; i < ARRAY_count(Class->array); i++)
  {
    array = &Class->array[i];

    write_long(TYPE_long(array->type));

    for (j = 0; j < array->ndim; j++)
    {
      p = array->dim[j];
      if (j == (array->ndim - 1))
        p = (-p);

      write_long(p);
    }
  }

  END_SECTION()
}


PRIVATE void output_string(void)
{
  START_SECTION()

  long i;
  SYMBOL *sym;

  #ifdef DEBUG
  printf("Section #13 : Chanes\n");
  #endif

  for (i = 0; i < TABLE_count(StringTable); i++)
  {
    sym = TABLE_get_symbol(StringTable, i);
    write_string(sym->name, sym->len);
  }

  END_SECTION()
}




PRIVATE void output_code()
{
  int i, ns;
  long n;
  FUNCTION *func;

  for (i = 0, ns = 14; i < ARRAY_count(Class->function); i++)
  {
    func = &Class->function[i];

    n = ARRAY_count(func->code);

    #ifdef DEBUG
    printf("Section #%d : Code (%d)\n", ns, n);
    ns++;
    #endif

    write_long(n * sizeof(ushort));
    write_buffer(func->code, n * sizeof(ushort));

    if (JOB->debug)
    {
      n = func->pos_line ? ARRAY_count(func->pos_line) : 0;

      #ifdef DEBUG
      printf("Section #%d : Line (%d)\n", ns, n);
      ns++;
      #endif

      write_long(n * sizeof(ushort));
      write_buffer(func->pos_line, n * sizeof(ushort));
    }
  }

  if (JOB->debug)
  {
    char path[MAX_PATH + 1];

    if (JOB->name[0] == '/')
    {
      strcpy(path, JOB->name);
    }
    else
    {
      getcwd(path, MAX_PATH);
      n = strlen(path);

      if (path[n - 1] != '/')
        strcpy(&path[n], "/");

      strcat(&path[n], JOB->name);
    }

    n = strlen(path);
    write_long(n + (n & 1));
    write_buffer(path, n);
    write_pad();

    #ifdef DEBUG
    printf("Section #%d : Source (%d)\n", ns, n);
    ns++;
    #endif

  }

}


PUBLIC char *OUTPUT_get_file(char *file)
{
  char *output;
  char *p;
  char *dir;
  char *name;

  dir = STR_copy(FILE_get_dir(file));
  name = STR_copy(FILE_get_name(file));

  for (p = name; *p; p++)
  {
    if (*p == '.')
    {
      *p = 0;
      break;
    }

    *p = toupper(*p);
  }

  output = FILE_cat(dir, ".gambas", NULL);
  mkdir(output, 0777);

  output = STR_copy(FILE_cat(dir, ".gambas", name, NULL));

  STR_free(dir);
  STR_free(name);

  return output;
}


PUBLIC void OUTPUT_do(void)
{
  char *name;

  output_init();

  /* La premire chane est toujours le nom de la classe */
  get_string(JOB->class->name, strlen(JOB->class->name));

  name = JOB->output;

  #ifdef DEBUG
  printf("Output to %s\n", name);
  #endif

  File = fopen(name, "w");
  if (!File)
    THROW("Cannot create file '%s'", name);

  Class = JOB->class;

  #ifdef DEBUG
  printf("pos = %lu\n", get_pos());
  #endif

  output_header();
  output_class();
  output_desc();
  output_constant();
  output_class_ref();
  output_unknown_ref();
  output_static();
  output_dynamic();
  output_event();
  output_method();
  output_param_local();
  output_array();
  output_string();
  output_code();

  fclose(File);

  output_exit();
}
