/***************************************************************************

  Object.c

  Object management routines

  (c) 2000-2001 Benot Minisini <gambas@freesurf.fr>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#define __OBJECT_C

#include "common.h"
#include "class.h"
#include "alloc.h"
#include "event.h"
#include "exec.h"

#include "Object.h"


PUBLIC void OBJECT_new(void **ptr, CLASS *class, char *name, OBJECT *parent)
{
  OBJECT_alloc(ptr, class, class->size);

  /* L'objet dmarre avec un compteur de rfrence  1
     Ce compteur devra tre remis  zro par un OBJECT_unref_keep()
     lorsque l'initialisation de l'objet sera termine
  */

  /*OBJECT_REF(*ptr, "OBJECT_new");*/

  #if DEBUG_EVENT
  printf("OBJECT_new: %s %p %d\n", class->name, *ptr, class->off_event);
  #endif

  OBJECT_attach(*ptr, parent, name);
}


PUBLIC void OBJECT_alloc(void **ptr, CLASS *class, size_t size)
{
  OBJECT *object;

  if (size < sizeof(OBJECT))
    ERROR_panic("OBJECT_alloc: size < %d", sizeof(OBJECT));

  ALLOC(&object, size, "OBJECT_alloc");
  memset(object, 0, size);

  object->class = class;
  object->ref = 1;
  /* Lorsque les initialisations sont termines, remettre
     le compteur de rfrences  zro */
  *ptr = object;
}


PUBLIC void OBJECT_detach(OBJECT *ob)
{
  CLASS *class = OBJECT_class(ob);
  OBJECT *parent;
  OBJECT_EVENT *ev;

  if (class->n_event <= 0)
    return;

  ev = (OBJECT_EVENT *)((char *)ob + class->off_event);

  /* Evite une rcursion infinie si la libration du parent provoque la libration de l'objet */
  parent = ev->parent;

  if (parent)
  {
    ev->parent = NULL;
    #if DEBUG_REF
      printf("OBJECT_detach : Detach (%s %p) from (%s %p)\n",
        ob->class->name, ob, parent->class->name, parent);
    #endif
    OBJECT_UNREF((void **)&parent, "OBJECT_detach");
  }
}


PUBLIC void OBJECT_attach(OBJECT *ob, OBJECT *parent, char *name)
{
  CLASS *class = OBJECT_class(ob);
  OBJECT_EVENT *ev;

  if (class->n_event == 0)
    return;

  OBJECT_detach(ob);

  ev = (OBJECT_EVENT *)((char *)ob + class->off_event);

  ev->parent = parent;
  #if DEBUG_REF
    printf("OBJECT_attach : Attach (%s %p) to (%s %p)\n",
      ob->class->name, ob, parent->class->name, parent);
  #endif
  OBJECT_REF(parent, "OBJECT_attach");

  EVENT_search(class, ev->event, name, parent);
}


PUBLIC void OBJECT_free(void **ptr)
{
  void *free_ptr = *ptr;
  CLASS *class = OBJECT_class(free_ptr);

  /* Il faut empcher EXEC_leave() de relibrer l'objet */
  /*((OBJECT *)free_ptr)->ref = 1;*/

  OBJECT_REF(free_ptr, "OBJECT_free");

  /* Excution de la mthode _free pour toutes les classes parentes
     puis pour la classe de l'objet */

  EXEC_special_inheritance(SPEC_FREE, class, (OBJECT *)free_ptr, 0, TRUE);

  OBJECT_detach(free_ptr);

  FREE(&free_ptr, "OBJECT_free");
}


PUBLIC boolean OBJECT_comp_value(VALUE *ob1, VALUE *ob2)
{
  if (ob1->type == T_NULL && ob2->type == T_NULL)
    return TRUE;
  else if (ob1->type == T_NULL)
    return ob2->_object.object == NULL;
  else if (ob2->type == T_NULL)
    return ob1->_object.object == NULL;
  else
    return (ob1->_object.object == ob2->_object.object);

  /*
  if (ob1->type == T_OBJECT)
    c1 = OBJECT_class(ob1->_object.object);
  else
    c1 = ob1->_object.class;

  if (ob2->type == T_OBJECT)
    c2 = OBJECT_class(ob2->_object.object);
  else
    c2 = ob1->_object.class;

  if (c1 != c2)
    THROW(E_TYPE);
  */
}



PUBLIC void OBJECT_release(CLASS *class, OBJECT *ob)
{
  int i;
  CLASS_VAR *var;
  int nelt;
  char *data;

  if (class->parent != NULL)
    OBJECT_release(class->parent, ob);

  if (CLASS_is_native(class))
    return;

#if DEBUG_ALLOC || DEBUG_MEMORY
  printf("> OBJECT_release %s %p\n", class->name, ob);
#endif

  if (ob == NULL)
  {
    var = class->load->stat;
    nelt = class->load->n_stat;
    data = class->stat;
  }
  else
  {
    var = class->load->dyn;
    nelt = class->load->n_dyn;
    data = (char *)ob;
  }

  for (i = 0; i < nelt; i++)
  {
#if DEBUG_ALLOC || DEBUG_MEMORY
    if (var->type.id == T_STRING || var->type.id == T_OBJECT)
    {
      printf("OBJECT_release: %s %p [%d] trying %p\n", class->name, ob, i, (*(void **)&data[var->pos]));
      fflush(NULL);
    }
#endif

    if (var->type.id == T_STRING)
      STRING_free((char **)&data[var->pos]);
    else if (var->type.id == T_OBJECT)
      OBJECT_unref((void **)&data[var->pos]);
    else if (var->type.id == T_VARIANT)
      VARIANT_free((VARIANT *)&data[var->pos]);
    else if (var->type.id == T_ARRAY)
      ARRAY_free_data((void *)&data[var->pos], (ARRAY_DESC *)class->load->array[var->type.value]);

    var++;
  }

#if DEBUG_ALLOC || DEBUG_MEMORY
  printf("< OBJECT_release %s %p\n", class->name, ob);
#endif
}

