/*

  trace.c
  
  Le debugger
  
*/


#define __TRACE_C

#include "common.h"
#include <unistd.h>

#include "error.h"
#include "type.h"
#include "debug.h"
#include "limit.h"

#include "class.h"
#include "exec.h"

#include "trace.h"


PUBLIC FUNCTION *TRACE_start_function = NULL;


PRIVATE TRACE_BREAK Breakpoint[MAX_BREAKPOINT];
PRIVATE int Reset_breakpoint = -1;

PRIVATE int Last_command;


PRIVATE int find_free_breakpoint(void)
{
  int i;
  TRACE_BREAK *tb;

  for (i = 1, tb = &Breakpoint[1]; i < MAX_BREAKPOINT; i++, tb++)
  {
    if (tb->addr == NULL)
      return i;
  }
  
  return 0;
}

PRIVATE boolean calc_func_addr(CLASS *class, ushort line, FUNCTION **function, PCODE **addr)
{
  int i;
  ushort pos, pos_after;
  FUNCTION *func;
  
  for (i = 0; i < class->load->n_func; i++)
  {
    func = &class->load->func[i];
    if (line >= func->line && line < (func->line + func->nline))
      break;
  }
  
  if (i >= class->load->n_func)
  {
    printf("Bad line number.\n");
    return TRUE;
  }

  for(;;)
  {
    pos = func->pos[line - func->line];
    pos_after = func->pos[line - func->line + 1];
    if (pos != pos_after)
      break;
      
    line++;
    if (line >= (func->line + func->nline))
      return TRUE;
  }  

  *function = func;  
  *addr = &func->code[pos];

  return FALSE;
}


PRIVATE boolean set_breakpoint(CLASS *class, ushort line, int *num)
{
  PCODE *addr;
  TRACE_BREAK *tb;
  int n;
  FUNCTION *func;
  
  if (CLASS_is_native(class) || !class->debug)
  {
    printf("No debugging information.\n");
    return TRUE;
  }

  if (calc_func_addr(class, line, &func, &addr))
    return TRUE;

  n = find_free_breakpoint();
  if (n == 0)
    return TRUE;
  
  tb = &Breakpoint[n];
  tb->func = func;
  
  if (PCODE_is_breakpoint(*addr))
  {
    printf("Breakpoint already set.\n");
    return TRUE;
  }

  tb->value = *addr;
  *addr = PCODE_BREAKPOINT(n);
  tb->addr = addr;
  
  if (num != NULL)
    *num = n;
  
  return FALSE;
}


PRIVATE boolean unset_breakpoint(CLASS *class, ushort line)
{
  TRACE_BREAK *tb;
  int i;
  FUNCTION *func;
  PCODE *addr;

  if (calc_func_addr(class, line, &func, &addr))
    return TRUE;
    
  for (i = 0; i < MAX_BREAKPOINT; i++)
  {
    tb = &Breakpoint[i];
    if (tb->addr == addr)
    {
      *(tb->addr) = tb->value;
      tb->addr = NULL;

      if (i == Reset_breakpoint)
        Reset_breakpoint = -1;

      return FALSE;
    }
  }

  printf("Unknown breakpoint.\n");
  return TRUE;
}


PRIVATE boolean reset_breakpoint(int num)
{
  TRACE_BREAK *tb;

  if (num < 0 || num >= MAX_BREAKPOINT)
    return TRUE;

  tb = &Breakpoint[num];

  if (tb->addr == NULL)
    return TRUE;

  *(tb->addr) = tb->value;

  EXEC_stop_next = TRUE;
  Reset_breakpoint = num;

  return FALSE;
}


PUBLIC ushort TRACE_get_current_line(void)
{
  int i;
  ushort pos = PC - FP->code;

  for (i = 0; i < (FP->nline - 1); i++)
  {
    if (pos >= FP->pos[i] && pos < FP->pos[i + 1])
      return i + FP->line;
  }

  return 0;
}


PRIVATE void break_on_next_line(void)
{
  ushort line;

  EXEC_stop_next_line = TRUE;

  if (CP == NULL)
  {
    EXEC_begin_line = NULL;
    EXEC_end_line = NULL;
  }
  else
  {
    line = TRACE_get_current_line();

    if (line == FP->nline)
      return;

    EXEC_begin_line = PC;
    EXEC_end_line = &FP->code[FP->pos[line + 1]];

    /*printf("break on next line : %d -> %d\n", PC - func->code, pos);*/
  }
}


PUBLIC void TRACE_main(void)
{
  static char last_command[256] = "";
  char cmd[256];
  ushort line;
  boolean loop;
  CLASS *class;

  do
  {
    loop = FALSE;
    printf("\n");

    if (CP == NULL)
    {
      line = 0;
      printf("(Not running)\n");
    }
    else
    {
      line = TRACE_get_current_line();
      printf("%s.%s.%d\n", CP->name, FP->name, FP->line + line);
    }

    printf("> ");
    fgets(cmd, sizeof(cmd), stdin);

    if (!strcmp(cmd, "\n"))
      strcpy(cmd, last_command);
    else
      strcpy(last_command, cmd);

    if (!strcmp(cmd, "q\n"))
    {
      abort();
    }
    else if (!strcmp(cmd, "g\n"))
    {
      Last_command = TC_GO;
      return;
    }
    else if (!strcmp(cmd, "s\n"))
    {
      Last_command = TC_STEP;
      break_on_next_line();
      EXEC_stop_function = NULL;
      return;
    }
    else if (!strcmp(cmd, "n\n"))
    {
      Last_command = TC_NEXT;
      break_on_next_line();
      EXEC_stop_function = FP;
      return;
    }
    else if (!strncmp(cmd, "+ ", 2))
    {
      char class_name[64];

      if (sscanf(cmd, "+ %64[^.].%hu", class_name, &line) != 2)
        printf("Syntax error.\n");
      else
      {
        class = CLASS_get(class_name);
        if (class == NULL)
          printf("Unknown class.\n");
        else if (!set_breakpoint(class, line, NULL))
          printf("OK\n");
      }

      loop = TRUE;
    }
    else if (!strncmp(cmd, "- ", 2))
    {
      char class_name[64];

      if (sscanf(cmd, "- %64[^.].%hu", class_name, &line) != 2)
        printf("Syntax error.\n");
      else
      {
        class = CLASS_get(class_name);
        if (class == NULL)
          printf("Unknown class.\n");
        else if (!unset_breakpoint(class, line))
          printf("OK\n");
      }

      loop = TRUE;
    }
    else
      loop = TRUE;
  }
  while (loop);
}


PUBLIC void TRACE_breakpoint(int num)
{
  if (reset_breakpoint(num))
    return;

  TRACE_main();
}


PUBLIC void TRACE_next(void)
{
  TRACE_BREAK *tb;

  if (Reset_breakpoint < 0)
    return;

  tb = &Breakpoint[Reset_breakpoint];
  *(tb->addr) = PCODE_BREAKPOINT(Reset_breakpoint);
  Reset_breakpoint = -1;
}


PUBLIC void TRACE_where(void)
{
  ushort line = 0;

  if (FP != NULL)
    line = TRACE_get_current_line();

  printf("%s.%s.", CP ? CP->name : NULL, FP ? FP->name : NULL);
  if (line == 0) printf("?"); else printf("%d", FP->line + line);
  printf(": ");
}
